import logging
import subprocess
import os
import shutil
import pwd
import json
from tempuscator.exceptions import BackupFileCorrupt, DirectoryNotEmpty, BackupCreateError
from typing import Union
from tempuscator.constants import (
    XBSTREAM_PATH,
    XTRABACKUP_PATH,
    SCP_PATH
)

_logger = logging.getLogger(__name__)


class BackupProcessor():
    """
    Xtrabackup backup processor class
    """

    def __init__(
            self,
            source: str,
            target: str,
            parallel: int = int(os.sysconf('SC_NPROCESSORS_ONLN')),
            force: bool = False,
            remove_backup: bool = False,
            user: Union[str, int] = os.getuid(),
            group: Union[str, int] = os.getgid(),
            logger_name: str = "Obfuscator",
            save_archive: str = None) -> None:
        self._log_level = _logger.getEffectiveLevel()
        self.target = target
        self.source = source
        self.force = force
        self.parallel = parallel
        self.user = user
        self.group = group
        self.remove_backup = remove_backup
        self.save_archive = save_archive
        if not os.path.isfile(self.source):
            raise FileNotFoundError(f"Backup {self.source} not found, or not regular file")
        if self.force:
            if os.path.exists(self.target):
                _logger.debug(f"Removing {self.target}")
                shutil.rmtree(path=self.target)
            if os.path.exists(self.save_archive):
                _logger.debug(f"Removing: {self.save_archive}")
                os.remove(self.save_archive)
        if os.path.isfile(path=self.target):
            raise FileExistsError(f"Destination {self.target} is regulara file, it should be empty dir or non existing path")
        if self.save_archive and os.path.exists(self.save_archive):
            raise FileExistsError(f"Destination {self.save_archive} already exists, not overwriting")
        if not os.path.exists(path=self.target):
            os.umask(0)
            os.makedirs(name=self.target, mode=0o751)
            if isinstance(self.user, str):
                user = pwd.getpwnam(self.user)
                os.chown(path=self.target, uid=user.pw_uid, gid=user.pw_gid)
        if os.path.isdir(self.target):
            empty = os.listdir(path=self.target)
            if len(empty) != 0:
                raise DirectoryNotEmpty(f"Directory {self.target} not empty")

    def __str__(self):
        return json.dumps(self.__dict__, indent=2)

    def extract(self, debug: bool = False) -> None:
        """
        Extract xtrabackup backup file
        """
        _logger.info(f"Extracting backup to {self.target}")
        cli = [XBSTREAM_PATH]
        cli.append("-x")
        cli.append("--directory")
        cli.append(self.target)
        cli.append("--decompress")
        cli.append(f"--decompress-threads={self.parallel}")
        cli.append("--parallel")
        cli.append(str(self.parallel))
        if debug:
            cli.append("--verbose")
        with open(self.source, 'r') as backup:
            _logger.debug(f"Executing: {' '.join(cli)}")
            extract = subprocess.Popen(cli, stdin=backup, user=self.user, group=self.group)
            extract.communicate()
            _logger.debug(f"Extract return code: {extract.returncode}")
            if not extract.returncode == 0:
                raise BackupFileCorrupt(f"File {self.source} looks like corruptted, try another")
        if self.remove_backup and extract.returncode == 0:
            log_msg = f"Removing {self.source}" if self._log_level <= 10 else "Removing source backup"
            _logger.log(self._log_level, log_msg)
            os.remove(self.source)

    def prepare(self, debug: bool = False) -> None:
        """
        Prepare extracted backup
        """
        output = None if debug else subprocess.DEVNULL
        _logger.info(f"Preparing restored backup in {self.target}")
        cli = [XTRABACKUP_PATH]
        cli.append("--prepare")
        cli.append("--target-dir")
        cli.append(self.target)
        _logger.debug(f"Executing: {' '.join(cli)}")
        prepare = subprocess.Popen(cli, stderr=output, user=self.user, group=self.group)
        prepare.wait()
        _logger.debug(f"Prepare exit code: {prepare.returncode}")

    def decompress(self, debug: bool = False) -> None:
        """
        Decompress extracted files
        """
        output = None if debug else subprocess.DEVNULL
        _logger.info("Decompressing files")
        cli = [XTRABACKUP_PATH]
        cli.append("--decompress")
        cli.append("--parallel")
        cli.append(str(self.parallel))
        cli.append("--remove-original")
        cli.append("--target-dir")
        cli.append(self.target)
        _logger.debug(f"Executing: {' '.join(cli)}")
        decompress = subprocess.Popen(cli, stderr=output, user=self.user, group=self.group)
        decompress.wait()
        _logger.debug(f"Decompress exit status: {decompress.returncode}")

    def create(
            self,
            dst: str,
            debug: bool = False,
            socket: str = "/var/lib/mysql/mysql.sock") -> None:
        """
        Create xtrabackup compressed archive (xbstream)
        """
        _logger.info("Creating xbstream archive")
        _logger.debug(f"Force: {self.force}")
        output = None if debug else subprocess.DEVNULL
        target_dir = "/tmp/xtrabackup_backupfiles/"
        if self.force and os.path.exists(dst):
            _logger.warning(f"Removing {dst}")
            os.remove(dst)
        cli = [XTRABACKUP_PATH]
        cli.append("--backup")
        cli.append("--target-dir")
        cli.append(target_dir)
        cli.append("--stream")
        cli.append("--compress")
        cli.append("--parallel")
        cli.append(str(self.parallel))
        cli.append("--compress-threads")
        cli.append(str(self.parallel))
        cli.append("--socket")
        cli.append(socket)
        cli.append(f"--datadir={self.target}")
        _logger.debug(f"Executing: {' '.join(cli)}")
        with open(dst, 'wb') as archive:
            backup = subprocess.Popen(cli, stdout=archive, stderr=output, user=self.user, group=self.group)
            backup.communicate()
        _logger.debug(f"Return code: {backup.returncode}")
        if backup.returncode > 0:
            raise BackupCreateError

    def cleanup_backup_files(self) -> None:
        """
        Remove not needed files and rotate certificates
        """
        _logger.info("Cleaning not needed files")
        files = [
            "auto.cnf",
            "backup-my.cnf",
            "ca-key.pem",
            "ca.pem",
            "client-cert.pem",
            "client-key.pem",
            "private_key.pem",
            "public_key.pem",
            "server-cert.pem",
            "server-key.pem",
            "xtrabackup_binlog_info",
            "xtrabackup_checkpoints",
            "xtrabackup_info",
            "xtrabackup_logfile",
            "xtrabackup_slave_info",
            "xtrabackup_tablespaces"
        ]
        for f in files:
            remove_file = os.path.join(self.target, f)
            _logger.debug(f"Removing: {remove_file}")
            if os.path.isfile(remove_file):
                os.remove(remove_file)

    def uploader(
            self,
            host: str,
            user: str,
            src: str,
            dst: str,
            progress: bool = False) -> None:
        """
        Upload new archive to destination server
        """
        _logger.info(f"Uploading file: {src} to {host}:{dst}")
        output = None if progress else subprocess.DEVNULL
        cli = [SCP_PATH]
        cli.append("-o")
        cli.append("UserKnownHostsFile=/dev/null")
        cli.append("-o")
        cli.append("StrictHostKeyChecking=no")
        cli.append("-o")
        cli.append("Compression=no")
        cli.append(src)
        cli.append(f"{user}@{host}:{dst}")
        _logger.debug(f"Executing: {' '.join(cli)}")
        upload = subprocess.Popen(cli, stdout=output, user=self.user, group=self.group)
        upload.communicate()

    def cleanup(self) -> None:
        """
        Remove target directory and all files inside
        """
        _logger.info("Cleaning up")
        _logger.debug(f"Removing: {self.target}")
        shutil.rmtree(self.target)



class DirectoryNotEmpty(Exception):
    """
    Directory not empty exception
    """


class BackupFileCorrupt(Exception):
    """
    Backup file corrupt exception
    """


class MysqldNotRunning(Exception):
    """
    Mysqld process not running exception
    """


class MysqlAccessDeniend(Exception):
    """
    Mysql access deniend exception
    """


class BackupCreateError(Exception):
    """
    Exception for failed backup create
    """


class MyCnfConfigError(Exception):
    """
    Mysql user config exception
    """


class MissingConfigSection(Exception):
    """
    Exception for missing config section in ini file
    """


class NotARoot(Exception):
    """
    Exception if running user not a root
    """



import logging
import subprocess
import psutil
import dataclasses
import sqlalchemy as db
import os
from typing import Union
from tempuscator.constants import MYSQLD_PATH


_logger = logging.getLogger(__name__)


@dataclasses.dataclass()
class MysqlData():

    datadir: str
    debug: bool
    user: Union[str, int] = dataclasses.field(default=os.getuid())
    group: Union[str, int] = dataclasses.field(default=os.getgid())
    mysql_user: str = dataclasses.field(default=os.environ["USER"])
    mysql_password: str = dataclasses.field(default=None)
    socket: str = dataclasses.field(init=False)
    pid: int = dataclasses.field(init=False, default=None)
    engine: db.Engine = dataclasses.field(init=False)
    running: bool = False
    conn_pool_size: int = dataclasses.field(default=4)

    def __post_init__(self) -> None:
        self.socket = os.path.join(self.datadir, "tempuscator.sock")
        url = ["mysql+pymysql://"]
        url.append(self.mysql_user)
        if self.mysql_password:
            url.append(":")
            url.append(self.mysql_password)
        url.append("@localhost/mysql?unix_socket=")
        url.append(self.socket)
        _logger.debug(f"Engine: {''.join(url)}")
        self.engine = db.create_engine(url="".join(url), echo=self.debug, pool_size=self.conn_pool_size)

    def __del__(self):
        """
        Destructor to stop mysqld on garbage collection
        """
        if self.running:
            self.stop()

    def start(self, skip_grants: bool = True) -> str:
        """
        Start mysqld service
        """
        _logger.info("Starting mysqld")
        pid_path = os.path.join(self.datadir, "tempuscator.pid")
        mysql_log = "/tmp/tempuscator.log"
        if not os.path.exists(mysql_log):
            with open(mysql_log, "w") as f:
                pass
        cli = [MYSQLD_PATH]
        if skip_grants:
            cli.append("--skip-grant-tables")
        cli.append("--datadir")
        cli.append(self.datadir)
        cli.append("--skip-networking")
        cli.append("--skip-name-resolve")
        cli.append("--skip-log-bin")
        cli.append("--socket")
        cli.append(self.socket)
        cli.append("--sync-binlog")
        cli.append("0")
        cli.append("--daemonize")
        cli.append("--pid-file")
        cli.append(pid_path)
        cli.append(f"--log-error={mysql_log}")
        cli.append("--sql-mode=")
        cli.append("--innodb-buffer-pool-instances=8")
        cli.append("--innodb-buffer-pool-size=6G")
        cli.append("--skip-innodb-doublewrite")
        cli.append("--innodb-flush-log-at-trx-commit=0")
        cli.append("--thread-pool-size=32")
        cli.append("--skip-performance-schema")
        cli.append("--skip-innodb-adaptive-hash-index")
        cli.append("--innodb-deadlock-detect=OFF")
        cli.append("--innodb-lock-wait-timeout=60")
        cli.append("--skip-innodb-buffer-pool-dump-at-shutdown")
        cli.append("--innodb-page-cleaners=8")
        cli.append("--innodb-log-buffer-size=128M")
        cli.append("--innodb-io-capacity=3000")
        cli.append("--innodb-io-capacity-max=6000")
        cli.append("--innodb-flush-neighbors=0")
        cli.append("--innodb-redo-log-capacity=4G")
        _logger.debug(f"Executing: {' '.join(cli)}")
        mysqld = subprocess.Popen(cli, stdout=subprocess.DEVNULL, user=self.user, group=self.group)
        mysqld.wait()
        with open(pid_path, 'r') as f:
            pid = f.read()
        self.pid = int(pid)
        self.running = True
        return self.socket

    def stop(self) -> None:
        """
        Stop mysqld service
        """
        if psutil.pid_exists(self.pid):
            _logger.info("Mysqld running, stopping")
            self.engine.dispose()
            proc = psutil.Process(pid=self.pid)
            proc.terminate()
            proc.wait()
            self.running = False
            return
        _logger.warning(f"Pid: {self.pid} doesn't exist")



import logging
import git
import os
import shutil
from typing import List


_logger = logging.getLogger(__name__)


class Scruber():

    def __init__(self, url: str, dst: str, sql_file: str) -> None:
        self.dst = dst
        if os.path.isdir(self.dst):
            _logger.debug(f"{dst} exists, pulling changes")
            repo = git.Repo(path=self.dst)
            repo.remotes.origin.pull()
        else:
            _logger.debug(f"{dst} doesn't exist, cloning repo")
            repo = git.Repo.clone_from(url=url, to_path=self.dst)
        self.source_file = os.path.join(dst, sql_file)
        if not os.path.exists(self.source_file):
            raise FileNotFoundError(f"{self.source_file} file not found")

    def __del__(self) -> None:
        if os.path.isdir(self.dst):
            shutil.rmtree(self.dst)

    def get_queries(self) -> List[str]:
        with open(self.source_file, "r") as f:
            return f.read().split("\n")[:-1]



import logging
import os
import configparser
import json
import inotify.adapters
import uuid
import time
import threading
from tempuscator.executor import Obfuscator
from tempuscator.engines import MysqlData
from tempuscator.swapper import SwapDirs
from tempuscator.exceptions import MissingConfigSection, NotARoot
from tempuscator.archiver import BackupProcessor
from tempuscator.repo import Scruber
from tempuscator.constants import CLOSE_WRITE_MASK

_logger = logging.getLogger(__name__)


class Watcher():

    def __init__(
            self,
            config: str,
            path: str,
            swapper: SwapDirs = None,
            mysql: MysqlData = None,
            obfuscator: Obfuscator = None,
            debug: bool = False) -> None:
        self.debug = debug
        self.path = path
        self.swapper = swapper
        self.mysql = mysql
        self.obfuscator = obfuscator
        if not os.path.isfile(config):
            raise FileNotFoundError(f"config file {config} not found!")
        if os.path.isfile(self.path):
            raise FileExistsError(f"{self.path} is regural file")
        if not os.path.exists(self.path):
            _logger.warning(f"Target directory {self.path} doesn't exist, creating")
            os.mkdir(path=self.path, mode=0o0750)
        raw_conf = configparser.RawConfigParser()
        with open(config, 'r') as f:
            raw_conf.read_file(f)
        if not raw_conf.has_section("obfuscator"):
            raise MissingConfigSection("Configuration missing section: obfuscator")
        self.conf: dict = raw_conf.__dict__.get("_sections")["obfuscator"]
        _logger.debug(f"config:\n{json.dumps(self.conf, indent=2)}")

    def watch_obfuscate(self) -> None:
        _logger.info("Starting obfuscator watcher")
        watcher = inotify.adapters.InotifyTree(path=self.path, mask=CLOSE_WRITE_MASK)
        for e in watcher.event_gen(yield_nones=False):
            (_, event, path, file) = e
            _logger.debug(f"received event: {e}")
            if "IN_CLOSE_WRITE" not in event:
                continue
            backup_file = os.path.join(path, file)
            _logger.info(f"Received: {backup_file}")
            self.__run_obfuscate(backup=backup_file)

    def watch(self, action: str) -> None:
        """
        Method to watch for uploaded files

        :param str action: Swap or obfuscate (obfuscate not implemented)

        :returns: None
        """
        actions = ["swap", "obfuscate"]
        if action not in actions:
            raise ValueError(f"action emust be one from: {' '.join(actions)}")
        _logger.info("Starting directory watcher")
        _logger.debug(f"Watching: {self.path}")
        watch = inotify.adapters.InotifyTree(path=self.path, mask=CLOSE_WRITE_MASK)
        for e in watch.event_gen(yield_nones=False):
            (_, event, path, file) = e
            _logger.debug(f"Received event: {e}")
            if "IN_CLOSE_WRITE" not in event:
                continue
            file = os.path.join(path, file)
            _logger.info(f"Received: {file}")
            self.__run_swap(backup=file)

    def _random_str(self) -> str:
        return uuid.uuid4().hex[:8]

    def __run_obfuscate(self, backup: str) -> None:
        start = time.perf_counter()
        repo_url = self.conf.get("repo")
        repo_dst = os.path.join("/tmp/", self._random_str())
        scrub_file = self.conf.get("scrub_sql")
        tmp_path = self.conf["tmp_path"] if "tmp_path" in self.conf.keys() else os.path.join("/tmp/", self._random_str())
        _logger.debug(f"Tmp path: {tmp_path}")
        processor = BackupProcessor(source=backup, target=tmp_path)
        _logger.debug(f"Backup procesor: {processor}")
        mysql = MysqlData(datadir=tmp_path, debug=self.debug)
        _logger.debug(f"Mysql data: {mysql}")
        scruber = Scruber(url=repo_url, dst=repo_dst, sql_file=scrub_file)
        obfuscator = Obfuscator(scrub=scruber)
        _logger.debug(f"Obfuscator: {obfuscator}")
        save_path = self.conf.get("save_path")
        try:
            processor.extract(debug=self.debug)
            processor.prepare(debug=self.debug)
            processor.cleanup_backup_files()
            socket = mysql.start()
            obfuscator.cleanup_system_users(engine=mysql.engine)
            obfuscator.change_system_user_password(engine=mysql.engine, user="root", empty=True)
            obfuscator.mask(engine=mysql.engine)
            processor.create(dst=save_path, socket=socket, debug=self.debug)
        finally:
            if socket:
                mysql.stop()
            processor.cleanup()
            _logger.debug(f"Removing: {backup}")
            os.remove(backup)
            end = time.perf_counter()
            execution_time = round((end - start)/60, 2)
            _logger.info(f"Execution time: {execution_time}")
        if "scp_host" in self.conf.keys():
            _logger.info("Uploading obfuscated backup")
            hosts = self.conf.get("scp_host").split(",")
            _logger.debug(f"Uploading to {hosts}")
            user = self.conf.get("ssh_user") if "ssh_user" in self.conf.keys() else os.environ["USER"]
            dst_path = self.conf.get('scp_path') if "scp_path" in self.conf.keys() else save_path
            u_thread = []
            for host in hosts:
                u_thread.append(threading.Thread(target=self.__run_upload, args=(processor, host, user, save_path, dst_path, )))
            for t in u_thread:
                t.start()
            for j in u_thread:
                j.join()

    def __run_upload(self, uploader: BackupProcessor, host: str, user: str, src: str, dst: str) -> None:
        """
        Threaded method for parallel uploader

        :param uploader: BackupProcessor class
        :param host: ip addres or hostname where to scp
        :param user: user name for ssh
        :param src: source file path
        :param dst: destination path were to put file

        :return: None
        """
        _logger.debug(f"Starting uploading to {host}")
        uploader.uploader(host=host, user=user, src=src, dst=dst)

    def __swap_checks(self) -> None:
        """
        Checks for swapping

        :raises NotARoot: user that runs this part of code must be a root
        """
        _logger.debug("Running prechecks for swapper")
        if os.getuid() != 0:
            raise NotARoot("User must be root")

    def __run_swap(self, backup) -> None:
        """
        Swap mysql directories

        :param str backup: Path to backup file

        :returns: None
        """
        self.__swap_checks()
        start = time.perf_counter()
        work_dir = os.path.join("/tmp/", self._random_str())
        swapper = SwapDirs(src_dir=work_dir)
        processor = BackupProcessor(source=backup, target=work_dir, user="mysql", group="mysql")
        mysql = MysqlData(datadir=work_dir, debug=self.debug, user="mysql", group="mysql")
        try:
            processor.extract(debug=self.debug)
            processor.prepare(debug=self.debug)
            mysql.start(skip_grants=False)
            swapper.update_users(engine=mysql.engine)
            mysql.stop()
            swapper.stop_mysqld()
            swapper.swap_dirs()
        finally:
            swapper.start_mysqld()
        stop = time.perf_counter()
        execution_time = round((stop - start)/60, 2)
        _logger.info(f"Program took: {execution_time} minutes")



import sqlalchemy as db
import logging

_logger = logging.getLogger(__name__)


def execute_query(engine: db.Engine, query: str, close: bool = False) -> None:
    """
    Executute raw query
    """
    with engine.connect() as conn:
        conn.execute(db.text(query))
        conn.commit()
    engine.dispose(close=close)



import logging
from logging.handlers import RotatingFileHandler
from tempuscator.constants import (
    LOG_FORMAT_DEFAULT,
    LOG_FORMAT_FILE_DEFAULT,
    LOG_FORMAT_DEBUG,
    LOG_FORMAT_FILE_DEBUG
)


def init_logger(name: str, level: str = "info", file: str = None) -> list:
    """
    Default logger initialization

    :param str name: Logger name for initialization
    :param str level: Root logging level, default info
    """
    log_levels = list(logging._nameToLevel.keys())[:-1]
    if level.upper() not in log_levels:
        raise ValueError(f"Log level {level} unknow")
    logger = logging.getLogger(name)
    set_level = logging.getLevelName(level.upper())
    log_format = logging.Formatter(LOG_FORMAT_DEBUG if set_level <= 10 else LOG_FORMAT_DEFAULT, style="{")
    logger.setLevel(set_level)
    con_logger = logging.StreamHandler()
    con_logger.setFormatter(log_format)
    con_logger.setLevel(set_level)
    logger.addHandler(con_logger)
    if file:
        log_format = logging.Formatter(LOG_FORMAT_FILE_DEBUG if set_level <= 10 else LOG_FORMAT_FILE_DEFAULT, style="{")
        rotating = RotatingFileHandler(filename=file, maxBytes=10240, mode='a', backupCount=3)
        rotating.setFormatter(log_format)
        rotating.setLevel(set_level)
        logger.addHandler(rotating)



# Executables

XBSTREAM_PATH = "/usr/bin/xbstream"
SCP_PATH = "/usr/bin/scp"
XTRABACKUP_PATH = "/usr/bin/xtrabackup"
MYSQLD_PATH = "/usr/sbin/mysqld"
PT_SHOW_GRANTS = "/usr/bin/pt-show-grants"
SYSTEMCTL_PATH = "/usr/bin/systemctl"
SSH_KEYSCAN_PATH = "/bin/ssh-keyscan"


# INotify masks
CLOSE_WRITE_MASK = 0x00000008

# Logging formats
LOG_FORMAT_DEFAULT = "{message}"
LOG_FORMAT_DEBUG = "[{levelname:^7}] - {name}: {message}"
LOG_FORMAT_FILE_DEFAULT = "{asctime}: " + LOG_FORMAT_DEFAULT
LOG_FORMAT_FILE_DEBUG = "{asctime} " + LOG_FORMAT_DEBUG



import dataclasses
import logging
import configparser
import sentry_sdk

_logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class Sentry():
    """
    Sentry data class for remote logging

    :param str dsn: Sentry dsn
    :param str env: Env name
    :return: None
    """
    dsn: str
    env: str

    def __post_init__(self) -> None:
        _logger.debug(f"Initializing sentry with dsn: {self.dsn} and env: {self.env}")
        sentry_sdk.init(
            dsn=self.dsn,
            environment=self.env
        )


def init_sentry(path: str) -> None:
    """
    Sentry initialization

    :param str path: path to config file containing sentry dsn and env
    :return: None
    """
    conf = configparser.RawConfigParser()
    with open(path, "r") as c_file:
        conf.read_file(c_file)
    if not conf.has_section("Sentry"):
        return
    Sentry(**conf["Sentry"])



import os
import time
import logging
from tempuscator.archiver import BackupProcessor
from tempuscator.executor import Obfuscator
from tempuscator.engines import MysqlData
from tempuscator.logger import init_logger
from tempuscator.arguments import obf_args, swap_args, notifier_args
from tempuscator.sentry import init_sentry
from tempuscator.swapper import SwapDirs
from tempuscator.base import Watcher


def obfuscator() -> None:
    """
    Entry point for tool
    """
    args = obf_args()
    log_level = "debug" if args.debug else args.log_level
    init_logger(name="tempuscator", level=log_level, )
    _logger = logging.getLogger(__name__)
    _logger.debug("Starting Obfuscator")
    _logger.debug(args)
    start = time.perf_counter()
    if os.path.isfile(args.config):
        _logger.debug(f"Initializing sentry from {args.config}")
        init_sentry(path=args.config)
    mysql = MysqlData(
        datadir=args.target_dir,
        debug=args.debug
    )
    backup = BackupProcessor(
        source=args.backup_file,
        target=mysql.datadir,
        force=args.force,
        parallel=args.parallel,
        remove_backup=args.remove_backup,
        save_archive=args.save_archive
    )
    obfuscator = Obfuscator(source=args.sql_file)
    backup.extract(debug=args.debug)
    xtrabackup_info = os.path.join(mysql.datadir, "xtrabackup_info")
    if not os.path.isfile(xtrabackup_info):
        backup.decompress(debug=args.debug)
    backup.prepare(debug=args.debug)
    backup.cleanup_backup_files()
    try:
        mysql.start()
        obfuscator.cleanup_system_users(engine=mysql.engine)
        obfuscator.change_system_user_password(engine=mysql.engine, user="root", empty=True)
        obfuscator.mask(engine=mysql.engine)
        backup.create(socket=mysql.socket, dst=args.save_archive, debug=args.debug)
    finally:
        mysql.stop()
        backup.cleanup()
    backup.uploader(
        host=args.host,
        user=args.ssh_user,
        src=args.save_archive,
        dst=args.scp_dst,
        progress=args.debug)
    stop = time.perf_counter()
    execution_time = round((stop - start)/60, 2)
    _logger.info(f"Program took: {execution_time} minutes")


def swapper() -> None:
    args = swap_args()
    log_level = "debug" if args.debug else args.log_level
    init_logger(name="tempuscator", level=log_level)
    _logger = logging.getLogger(__name__)
    _logger.debug(f"ARGS: {args}")
    start = time.perf_counter()
    if os.path.isfile(args.config):
        _logger.debug(f"Initializing sentry from {args.config}")
        init_sentry(path=args.config)
    swapper = SwapDirs(
            src_dir=args.extract_dir,
            user=args.mysql_user,
            password=args.mysql_password,
            backup=args.backup)
    _logger.debug(f"Swapper: {swapper}")
    backup = BackupProcessor(
            source=args.backup_file,
            target=swapper.src_dir,
            force=args.force,
            user=args.user,
            group=args.group,
            logger_name="Swapper",
            remove_backup=args.remove_backup)
    _logger.debug(f"Backup processor: {backup}")
    updated_data = MysqlData(
            datadir=swapper.src_dir,
            debug=args.debug,
            user=args.user,
            group=args.group)
    _logger.debug(f"Mysql data: {updated_data}")
    backup.extract(debug=args.debug)
    backup.decompress(debug=args.debug)
    backup.prepare(debug=args.debug)
    try:
        updated_data.start(skip_grants=False)
        swapper.update_users(updated_data.engine)
    finally:
        updated_data.stop()
    try:
        swapper.stop_mysqld()
        swapper.swap_dirs()
    finally:
        swapper.start_mysqld()
    stop = time.perf_counter()
    execution_time = round((stop - start)/60, 2)
    _logger.info(f"Program took: {execution_time} minutes")


def mysql_obf_watcher() -> None:
    args = notifier_args()
    if args.log_file:
        init_logger(name="tempuscator", level=args.log_level, file=args.log_file)
    else:
        init_logger(name="tempuscator", level=args.log_level)
    _logger = logging.getLogger(__name__)
    _logger.info("Starting inotify")
    _logger.debug(f"ARGS: {args}")
    if os.path.isfile(args.config):
        _logger.debug(f"Initializing sentry from {args.config}")
        init_sentry(path=args.config)
    listener = Watcher(config=args.conf_action, path=args.watch_dir, debug=args.debug)
    listener.watch_obfuscate()


def mysql_swap_watch() -> None:
    """
    Cli entry point for mysql directory wacher
    """
    args = notifier_args()
    if args.log_file:
        init_logger(name="tempuscator", level=args.log_level, file=args.log_file)
    else:
        init_logger(name="tempuscator", level=args.log_level)
    _logger = logging.getLogger(__name__)
    _logger.info("Starting inotify")
    _logger.debug(f"ARGS: {args}")
    if os.path.isfile(args.config):
        _logger.debug(f"Initializing sentry from {args.config}")
        init_sentry(path=args.config)
    listener = Watcher(config=args.conf_action, path=args.watch_dir, debug=args.debug)
    listener.watch(action="swap")



import argparse
import os
import logging


def base_args() -> argparse.ArgumentParser:
    """
    Base arguments
    """
    c_user_home = os.path.expanduser("~")
    default_config = os.path.join(c_user_home, ".tempuscator")
    log_levels = [level.lower() for level in logging._nameToLevel.keys()][:-1]
    args = argparse.ArgumentParser()
    base = args.add_argument_group(title="Base", description="Base parameters")
    base.add_argument(
        "--log-level",
        choices=log_levels,
        help=f"Log level from: {' '.join(log_levels)}, default %(default)s",
        default="info"
    )
    base.add_argument(
        "--log-file",
        help="Path to log file",
        type=str
    )
    base.add_argument(
        "--debug",
        help="enable debuging",
        action="store_true"
    )
    base.add_argument(
        "--force",
        help="Force remove and recreate target directory",
        action="store_true"
    )
    base.add_argument(
        "-c",
        "--config",
        help="Path to config file, default: %(default)s",
        default=default_config
    )
    return args


def obf_args() -> argparse.Namespace:
    """
    Arguments for obfuscation
    """
    args = base_args()
    ssh_args = args.add_argument_group(title="SSH", description="Arguments for uploading file")
    archiver = args.add_argument_group(title="Archiver")
    obfuscator = args.add_argument_group(title="Obfuscator")
    archiver.add_argument(
        "-b",
        "--backup-file",
        help="Path to file",
        required=True
    )
    archiver.add_argument(
        "--remove-backup",
        help="Remove backup after extracting",
        action="store_true"
    )
    archiver.add_argument(
        "--target-dir",
        help="Where to extract files",
        default="/tmp/obfuscation"
    )
    archiver.add_argument(
        "--save-archive",
        help="Path were to save obfuscated archive",
        required=True
    )
    archiver.add_argument(
        "-p",
        "--parallel",
        help="Parallel parameter for xtrabackup",
        default=4
    )
    obfuscator.add_argument(
        "--sql-file",
        help="Patgh to sql file",
        required=True
    )
    ssh_args.add_argument(
        "--host",
        type=str,
        help="Address of the server were to upload file"
    )
    ssh_args.add_argument(
        "--ssh-user",
        type=str,
        help="Username for ssh connection"
    )
    ssh_args.add_argument(
        "--scp-dst",
        type=str,
        help="File path were to put file"
    )
    return args.parse_args()


def swap_args() -> argparse.Namespace:
    """
    Arguments for swapper
    """
    args = base_args()
    archiver = args.add_argument_group(title="Archiver")
    swapper = args.add_argument_group(title="Swapper")
    mysql = args.add_argument_group(title="Mysql")
    mysql.add_argument(
        "--mysql-user",
        type=str,
        help="User for connecting to mysql"
    )
    mysql.add_argument(
        "--mysql-password",
        type=str,
        help="Password for connecting to mysql"
    )
    archiver.add_argument(
        "--user",
        help="User to which change permissions",
        type=str,
        default="mysql"
    )
    archiver.add_argument(
        "--group",
        help="Group for permission change",
        type=str,
        default="mysql"
    )
    archiver.add_argument(
        "-b",
        "--backup-file",
        help="Path to file",
        type=str,
        required=True
    )
    archiver.add_argument(
        "--extract-dir",
        help="Where to extract files, default: %(default)s",
        type=str,
        default="/tmp/swapper",
    )
    archiver.add_argument(
        "--remove-backup",
        help="Remove backup after extracting",
        action="store_true"
    )
    swapper.add_argument(
        "--backup",
        help="Leave backup directory of previuos mysql version",
        action="store_true"
    )
    return args.parse_args()


def notifier_args() -> argparse.Namespace:
    args = base_args()
    notifier = args.add_argument_group(title="Notifier")
    notifier.add_argument(
        "--watch-dir",
        help="Directory to watch, default: %(default)s",
        type=str,
        default="/tmp/notifier"
    )
    notifier.add_argument(
        "--action",
        help="Action to call on IN_CLOSE_WRITE",
        required=False,
        choices=["obfuscate", "swap"]
    )
    notifier.add_argument(
        "--conf-action",
        help="Action config file",
        type=str,
        required=True
    )
    return args.parse_args()



import dataclasses
import psutil
import os
import subprocess
from tempuscator.exceptions import MysqldNotRunning, MysqlAccessDeniend, MyCnfConfigError
from tempuscator.helpers import execute_query
from typing import List
import sqlalchemy as db
import logging
import datetime
import shutil
import configparser
from tempuscator.constants import (
    PT_SHOW_GRANTS,
    SYSTEMCTL_PATH
)

_logger = logging.getLogger(__name__)


@dataclasses.dataclass
class SwapDirs():
    src_dir: str
    user: str = dataclasses.field(default=None)
    password: str = dataclasses.field(default=None, repr=False)
    grants: List[str] = dataclasses.field(init=False, default_factory=list, repr=False)
    dst_dir: str = dataclasses.field(default="/var/lib/mysql")
    backup: bool = dataclasses.field(default=False)
    mysqld_running: bool = dataclasses.field(default=False)

    def __post_init__(self):
        if not os.path.exists(PT_SHOW_GRANTS):
            raise FileNotFoundError(f"{PT_SHOW_GRANTS} not found!")
        self.mysqld_running = "mysqld" in (p.name() for p in psutil.process_iter())
        if not self.mysqld_running:
            raise MysqldNotRunning("Mysqld not running")
        u_my_cnf = os.path.join(os.environ["HOME"], ".my.cnf")
        if os.path.exists(u_my_cnf):
            _logger.debug(f"Parsing {u_my_cnf}")
            u_conf = configparser.ConfigParser()
            with open(u_my_cnf, "r") as my_cnf:
                u_conf.readfp(my_cnf)
            if not u_conf.has_section("client"):
                raise MyCnfConfigError("User config doesn't have section client")
            if u_conf.has_option("client", "user"):
                self.user = u_conf.get("client", "user")
                _logger.debug(f"Mysql user: {self.user}")
            if u_conf.has_option("client", "password"):
                self.password = u_conf.get("client", "password")
        cli = [PT_SHOW_GRANTS]
        cli.append("--database")
        cli.append("mysql")
        cli.append("--host")
        cli.append("localhost")
        cli.append("--ignore")
        cli.append("root@localhost")
        if self.user:
            cli.append("--user")
            cli.append(self.user)
            cli.append("--password")
            cli.append(self.password)
        exec = subprocess.Popen(cli, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, err = exec.communicate()
        if err:
            raise MysqlAccessDeniend("Unable to connect to mysql")
        perms = out.decode().split("\n")[:-1]
        self.grants = [p for p in perms if not p.startswith("--")]

    def update_users(self, engine: db.Engine) -> None:
        _logger.info("Adding users")
        for grant in self.grants:
            execute_query(engine=engine, query=grant)
        _logger.info("Updating root password")
        query = db.text("ALTER USER {user}@localhost IDENTIFIED WITH caching_sha2_password BY '{password}'".format(
            user=self.user,
            password=self.password))
        with engine.connect() as conn:
            conn.execute(query)
            conn.commit()
        engine.dispose()

    def stop_mysqld(self) -> None:
        _logger.info("Stopping system Mysqld")
        if not self.mysqld_running:
            raise MysqldNotRunning("System mysqld not running")
        cli = [SYSTEMCTL_PATH]
        cli.append("stop")
        cli.append("mysqld")
        exec = subprocess.Popen(cli)
        exec.communicate()
        if exec.returncode == 0:
            self.mysqld_running = False

    def start_mysqld(self) -> None:
        _logger.info("Starting system Mysqld")
        if self.mysqld_running:
            return
        cli = [SYSTEMCTL_PATH]
        cli.append("start")
        cli.append("mysqld")
        exec = subprocess.Popen(cli)
        exec.communicate()
        if exec.returncode == 0:
            self.mysqld_running = True

    def swap_dirs(self) -> None:
        _logger.info("Swapping system Mysqld directories")
        if self.backup:
            _logger.info("Saving old directory")
            self.__backup_old_mysqld_directory()
        if os.path.exists(self.dst_dir):
            _logger.debug(f"Removing {self.dst_dir}")
            shutil.rmtree(path=self.dst_dir)
        _logger.debug(f"Moving {self.src_dir} to {self.dst_dir}")
        os.rename(self.src_dir, self.dst_dir)

    def __backup_old_mysqld_directory(self) -> None:
        _logger.debug("Creating backup of old mysqld directory")
        backup_suffix = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        new_name = f"{self.dst_dir}-{backup_suffix}"
        _logger.debug(f"New name: {new_name}")
        os.rename(self.dst_dir, new_name)



import sqlalchemy as db
import logging
import threading
from tempuscator.helpers import execute_query
from tempuscator.repo import Scruber
import json

_logger = logging.getLogger(__name__)


class Obfuscator():

    def __init__(self, scrub: Scruber) -> None:
        self.queries = scrub.get_queries()

    def __str__(self) -> str:
        return json.dumps(self.__dict__)

    def change_system_user_password(self, user: str, engine: db.Engine, empty: bool = False) -> None:
        password = ""
        if not empty:
            password = "somepassword"
        meta = db.MetaData()
        meta.reflect(bind=engine)
        USER = meta.tables["user"]
        with engine.connect() as conn:
            query = USER.update().where(USER.c.User == user).values(authentication_string=password)
            conn.execute(query)
            conn.commit()

    def cleanup_system_users(self, engine: db.Engine) -> None:
        _logger.info("Cleaning up users")
        meta = db.MetaData()
        meta.reflect(bind=engine)
        USER = meta.tables["user"]
        query = db.delete(
            USER
        ).filter(
            db.not_(
                USER.c.User.in_([
                    "root",
                    "mysql.sys",
                    "mysql.infoschema",
                    "mysql.session"])
                )
            )
        with engine.connect() as conn:
            conn.execute(query)
            conn.commit()

    def mask(self, engine: db.Engine) -> None:
        _logger.info("Executing masking queries")
        threads = []
        for q in self.queries:
            threads.append(threading.Thread(target=execute_query, args=(engine, q, )))
        for t in threads:
            t.start()
        for j in threads:
            j.join()
